
/*
 * Code produces four different reports in any combination.
 *	0: show airport which minimizes maximum distance
 *	1: list of all airport-to-airport distances
 *	2: show centralmost airports for various exponents
 *	4: show average distances from each airport.
 *
 * Report 0 takes only two lines and is always printed.
 * Report 1 produces 700,000+ lines so is time-consuming
 *	and may be useless without further processing.
 * Report 2 answers first question posed in SDMB thread.
 * Report 4 answers second question posed in SDMB thread.
 *
 * Enter reports you want on command line; for examples:
 *	airdist 2 -- output Report 2
 *	airdist 6 -- output Reports 2 and 4 (6 = 2+4)
 *	airdist 7 -- output all Reports (7 = 1+2+4)
 *	airdist	-- defaults to airdist 6 (Reports 2 and 4)
 */

#include	<stdlib.h>
#include	<stdio.h>
#include	<math.h>

#define EXIT_GOOD       0
#define EXIT_SWCONF     24      /* software misconfigured */
#define EXIT_USAGE      64      /* command line usage error */

/* Code knows the true size of Port[] so the
 * Next number doesn't matter as long as it's big enough
 */
#define	NUMPORT	1221

/* Giant array could be avoided, but shouldn't matter
 * with the giant memories of today's computers.
 */
double Logdist[NUMPORT][NUMPORT];

struct port {
	char    *n_icao, *n_iata;
	char    *n_port, *n_town, *n_country;
	double  altitude;
	double  latitu, longit;
} Port[] = {
#include	"airports.h"
};

double Porthelp[10][2];

#define	J_PI  3.1415926535898

/* We use only an average Earth radius for calculations */
double Radnet = 6371;	/* Earth radius in km */
/*
 * double Radpolar = 6356.9;
 * double Radequat = 6378.5;
 */

double t_dist(double to[2], double fro[2])
{
        double	z = cos(to[0]) * cos(fro[1] - to[1]);
        double	y = sin(to[0]);
        y = - sqrt(y*y + z*z) * cos(atan2(y, z) - fro[0]);
        return 6371 * (3.14159 / 2 + atan2(y, sqrt(1 - y*y)));
}

#define	d2r(x)	((x) * J_PI / 180)

/* Calculate distance between two places */
double getd2(int ix1, int ix2)
{
	double long1, lati1, long2, lati2;

	Porthelp[0][1] = d2r(Port[ix1].longit);
	Porthelp[0][0] = d2r(Port[ix1].latitu);
	Porthelp[1][1] = d2r(Port[ix2].longit);
	Porthelp[1][0] = d2r(Port[ix2].latitu);
	return t_dist(Porthelp[0], Porthelp[1]);
}

int	Reportmode = 6;

int main(int argc, char **argv)
{
	int	siz, ix1, ix2;
	double	dista, distt;
	double	expon, tsum, winsum;
	double	longest;
	int	keepix2 = 0;
	int	winix = 0;

	if (argv[1])
		Reportmode = atoi(argv[1]);
	fprintf(stderr, "Reportmode = %d\n", Reportmode);
	siz = sizeof Port / sizeof *Port;
	fprintf(stderr, "Processing %d airports\n", siz);
	if (siz > NUMPORT) {
		fprintf(stderr, "Recompile with larger Numport\n");
		exit(EXIT_SWCONF);
	}

	for (ix1 = 0; ix1 < siz; ix1++)
	for (ix2 = ix1 + 1; ix2 < siz; ix2++) {
		dista = getd2(ix1, ix2);
		Logdist[ix1][ix2] =
			Logdist[ix2][ix1] = log(dista);
		if (Reportmode & 1)
		printf(" %8.1f is Distance from %s (%s,%s) to %s (%s,%s)\n",
			dista,
			Port[ix1].n_iata,
			Port[ix1].n_town,
			Port[ix1].n_country,
			Port[ix2].n_iata,
			Port[ix2].n_town,
			Port[ix2].n_country);
	}

	winsum = 99999;
	for (ix1 = 0; ix1 < siz; ix1++) {
		longest = 0;
		tsum = 0;
		for (ix2 = 0; ix2 < siz; ix2++) {
			if (ix2 != ix1) {
				distt = exp(1 * Logdist[ix1][ix2]);
				tsum += distt;
				if (distt > longest) {
					longest = distt;
				} 
			}
		}
		if (longest < winsum) {
			winsum = longest;
			winix = ix1;
		}
		if (Reportmode & 4)
		printf("%f is average distance from %s (%s,%s)\n",
			tsum / (siz - 1),
			Port[ix1].n_iata,
			Port[ix1].n_town,
			Port[ix1].n_country);
	}
	printf("Maximum distance is minimized with %f at %s (%s,%s)\n",
			winsum,
			Port[winix].n_iata,
			Port[winix].n_town, Port[winix].n_country);
	longest = 0;
	for (ix2 = 0; ix2 < siz; ix2++) {
		if (ix2 != winix) {
			tsum = exp(1 * Logdist[winix][ix2]);
			if (tsum > longest) {
				longest = tsum;
				keepix2 = ix2;
			} 
		}
	}
	printf(" ... to %s (%s,%s)\n",
			Port[keepix2].n_iata,
			Port[keepix2].n_town, Port[keepix2].n_country);

	if (Reportmode & 2)
	for (expon = .10; expon < 3.01; expon += .10) {
		winsum = 99999 * siz;
		for (ix1 = 0; ix1 < siz; ix1++) {
			tsum = 0;
			for (ix2 = 0; ix2 < siz; ix2++)
				if (ix2 != ix1) {
					tsum += exp(expon * Logdist[ix1][ix2]);
				}
			if (tsum < winsum) {
				winsum = tsum;
				winix = ix1;
			}
		}
		printf("Centralest airport (Min sum dist^%4.2f) is %s (%s,%s)\n",
			expon, Port[winix].n_iata,
			Port[winix].n_town, Port[winix].n_country);
			// winsum / (siz - 1));
	}
	exit(EXIT_GOOD);
}
