
/* Copyright (c) 2008 James Dow Allen.  All Rights Reserved; see License. */
/*
 * Hash table handling for fixed-length keys, using Cuckoo
 */

#define		MAIN_MODULE
#include	"js_simple.h"

/* Keys and Bgroups need NOT be same size */
typedef Uint64 Key;
#ifdef  BIGENTRY
typedef Uint64 Bgroup;
#else
typedef Uint32 Bgroup;
#endif

#include	"bitaccess.h"
#include	"cuckhash.h"
#include	"genhash.h"

void htab_sstat(struct htab_tab *csp)
{
	fprintf(stderr,
		"Stats from cuckoo: %ld bytes in %d parts\n",
			(long)csp->cu_bytes,
			csp->cu_numparts);
	fprintf(stderr,
  		"     %ld adds  %ld dels  %ld reloc  %ld finds  %ld probes\n",
			csp->cu_numadd, csp->cu_numdel,
			csp->cu_numrelo, csp->cu_find, csp->cu_probe);
	fprintf(stderr,
  		"  Scrambler:  prime= %ld fm= %ld im= %ld ig= %d\n",
			csp->cu_prim, csp->cu_fm, csp->cu_im, csp->cu_igb);
	fprintf(stderr,
  		"    Table:  minsize= %d pay= %d bentry= %d\n",
			csp->cu_mintab, csp->cu_paysz, csp->cu_bentry);
}

/* Allocate (with siz entries) and clear ix'th compartment in *csp */
static void alloctab(struct htab_tab *csp, Uint ix, size_t siz)
{
	Uint	asiz;
	size_t	numbytes;

	csp->cu_nent[ix] = siz;
	asiz = ALLOSZ(siz * csp->cu_bentry);
	numbytes = asiz * sizeof(Bgroup);
	csp->cu_ddd[ix] = mustmalloc(numbytes, "allocating cuckoo compartment");
	csp->cu_bytes += numbytes;
	memset(csp->cu_ddd[ix], 0, numbytes);
}

static int c_nextg(struct htab_tab *csp, struct htab_trav *wp, Key *gp)
{
	Key	g;
	Bgroup	tval;
	int	rem;
	unsigned int bentry = csp->cu_bentry;
	unsigned int paysize = csp->cu_paysz;
	Bgroup *cbp = csp->cu_ddd[wp->wktab];
	size_t cbsize = csp->cu_nent[wp->wktab];

	if (wp->wkreset) {
		wp->wkindx = 0;
		wp->wkreset = 0;
		goto sttab;
	}
	do {
		if (++(wp->wkindx) >= (long)cbsize)
			return 0;
		sttab:
		rem = wp->wkindx;
		tval = Jgetslot(cbp, rem, bentry);
		g = tval >> paysize;
	} while (! g);
	g = (g - 1) * cbsize + rem;
	*gp = SCRAM(g, csp->cu_igb, csp->cu_im, csp->cu_prim);
	return  1 + (tval & ((1 << paysize) - 1));
}

/* Evacuate compartment zero, renumbering compartments
 * so the old 1 becomes 0, 2 becomes 1, ...
 *  and cu_numparts becomes cu_numparts - 1.
 * (which is of course the last of cu_numparts compartments
 *   numbered 0, 1, ...)
 */
static void cuck_movecpt(struct htab_tab *csp)
{
	Bgroup	*cdtp;
	size_t	ctsize;
	int	retc, findres, ix;
	struct htab_trav ws;
	Key	g;

	csp->cu_movecpt = 1;
	cdtp = csp->cu_ddd[0];
	ctsize = csp->cu_nent[0];
	for (ix = 0; ix < csp->cu_numparts; ix++) {
		csp->cu_ddd[ix] = csp->cu_ddd[ix + 1];
		csp->cu_nent[ix] = csp->cu_nent[ix + 1];
	}
	csp->cu_ddd[ix] = cdtp;
	csp->cu_nent[ix] = ctsize;
	fprintf(stderr, "cuck_movecpt moving %d elements\n", ctsize);
	ws.wkreset = 1;
	ws.wktab = ix;
	while (retc = c_nextg(csp, &ws, &g)) {
		findres = htab_find(csp, g, retc - 1, HCreate, (int *)0);
		if (findres < 0) {
			fprintf(stderr, "Unexpected duplicate during cuck_movecpt\n");
			exit(EXIT_SOFTWARE);
		}
		csp->cu_numadd -= 1; /* correct count */
	}
	free(csp->cu_ddd[ix]);
	csp->cu_ddd[ix] = 0;
	csp->cu_nent[ix] = 0;
	csp->cu_movecpt = 0;
	return;
}

Uint64 htab_maxkey(struct htab_tab *csp)
{
	return ((Uint64) csp->cu_prim << csp->cu_igb) - 1;
}

Uint64 htab_maxtag(struct htab_tab *csp)
{
	return 0; /* no support */
}

Uint htab_maxpoint(struct htab_tab *csp)
{
	return 0; /* no support */
}

int htab_find(struct htab_tab *csp, Key g, int val, htab_op hop, int *tagp)
{
	struct {
		Bgroup	quot;
		Uint	rem;
	} res[cuck_MAXPARTS];
	Uint	rem, primsz, pmask;
	Uint	bentry, paysize;
	int	cuckix;
	Bgroup	quo;
	Bgroup	tval;
	int	maxrty, rty, freet = -1;

	bentry = csp->cu_bentry;
	paysize = csp->cu_paysz;
	pmask = (1 << paysize) - 1;
	if (val < 0 || val > (int)pmask)
		return -1;

	csp->cu_find += 1;

	g = SCRAM(g, csp->cu_igb, csp->cu_fm, csp->cu_prim);

	for (cuckix = 0; cuckix < csp->cu_numparts; cuckix++) {
		primsz = csp->cu_nent[cuckix];
		quo = res[cuckix].quot = g / primsz + 1;
		rem = res[cuckix].rem = g % primsz;
		tval = Jgetslot(csp->cu_ddd[cuckix], rem, bentry);
		csp->cu_probe += 1;
		if (tval >> paysize == quo) {
			tval &= pmask;
			switch (hop) {
			case Hadd:
			case HCadd:
				val += tval;
				if (val > (int)pmask)
					goto norep;
				else
					goto dorep;
			case HLrep:
			case HLmak:
				if (val >= (int)tval)
					goto norep;
				/* FALLTHRU */
			case HArep:
			case HAmak:
			dorep:
				Jsetpay(csp->cu_ddd[cuckix], rem, val,
					paysize, bentry);
				return htab_OLDE(tval);
			case HHrep:
			case HHmak:
				if (val > (int)tval)
					goto dorep;
				/* FALLTHRU */
			case HLook:
			case HCreate:
			norep:
				return htab_OLDE(tval);
			case HDelete:
				Jsetslot(csp->cu_ddd[cuckix], rem, 0, bentry);
				csp->cu_numdel += 1;
				return htab_OLDE(tval);
			default:
				fprintf(stderr, "Unknown htab_op %d\n", hop);
				exit(EXIT_INVOKE);
			}
		}
		if (freet < 0 && tval == 0)
			freet = cuckix;
	}
	if (hop < HCreate)
		return 0;
	csp->cu_numadd += 1;
	if (freet >= 0) {
		cuckix = freet;
		rem = res[cuckix].rem;
		quo = res[cuckix].quot;
    finwrit2:
		Jsetslot(csp->cu_ddd[cuckix], rem,
			(Bgroup)quo << paysize | val, bentry);
		return 0;
	}

	freet = SIMP_RIX(csp->cu_numparts);
	if (csp->cu_numparts < 3)
		maxrty = 14;
	else if (csp->cu_numparts < 4)
		maxrty = 10;
	else
		maxrty = 7;
	if (csp->cu_numparts >= csp->cu_maxparts)
		maxrty += maxrty + 30;
	if (csp->cu_movecpt)
		maxrty += maxrty + 30;
	for (rty = 1; rty < maxrty; rty++) {
		tval = Jgetslot(csp->cu_ddd[freet], res[freet].rem, bentry);
		csp->cu_probe += 1;
		g = ((tval >> paysize) - 1)
			* (Key)csp->cu_nent[freet] + res[freet].rem;
		rem = res[freet].rem;
		quo = res[freet].quot;
		Jsetslot(csp->cu_ddd[freet], rem,
				(Bgroup)quo << paysize | val, bentry);
		csp->cu_numrelo += 1;
		val = tval & pmask;
		for (cuckix = csp->cu_numparts - 1; cuckix >= 0; cuckix--)
		if (cuckix != freet) {
			primsz = csp->cu_nent[cuckix];
			quo = res[cuckix].quot = g / primsz + 1;
			rem = res[cuckix].rem = g % primsz;
			tval = Jgetslot(csp->cu_ddd[cuckix], rem, bentry);
			csp->cu_probe += 1;
			if (tval == 0) {
				goto finwrit2;
			}
		}
		cuckix = SIMP_RIX(csp->cu_numparts - 1);
		freet = cuckix + (cuckix >= freet);
	}
	cuckix = csp->cu_numparts;
	if (cuckix > csp->cu_maxparts || csp->cu_movecpt) {
		fprintf(stderr, "Cuckoo exhaustion (%d %d)\n",
			cuckix, csp->cu_movecpt);
		exit(EXIT_EXCEED);
	}
	primsz = csp->cu_nent[cuckix - 1];
	primsz += primsz / 8;
	primsz = nxtprime(primsz);
	alloctab(csp, cuckix, primsz);
	res[cuckix].quot = g / primsz + 1;
	res[cuckix].rem = g % primsz;
	rem = res[cuckix].rem;
	quo = res[cuckix].quot;
	Jsetslot(csp->cu_ddd[cuckix], rem,
			(Bgroup)quo << paysize | val, bentry);
	if (cuckix == csp->cu_maxparts) {
		cuck_movecpt(csp);
	} else {
		csp->cu_numparts = cuckix + 1;
	}
	return 0;
}

/* Create a cuckoo table with a minimum of minnt compartments
 * maximimum of maxnt compartments, each part at least minsz entries
 */
struct htab_tab *cuck_create(size_t minsz, size_t minnt, size_t maxnt,
		struct cuck_revdesc *revdp)
{
	struct htab_tab *csp;
	size_t	primsz;
	int	ix;

	if (maxnt > cuck_MAXPARTS)
		maxnt = cuck_MAXPARTS;
	if (minnt > maxnt)
		return 0;
	if (minsz < 3)
		minsz = 3;
	if (revdp->re_bentry > 63) {
		fprintf(stderr, "Bentry = %d max = %d\n",
			revdp->re_bentry, 63);
		return 0;
	}
	if (revdp->re_bentry > BITS_BG - 1) {
		fprintf(stderr, "Bentry = %d recompile with -DBIGENTRY\n",
			revdp->re_bentry);
		return 0;
	}
	csp = mustmalloc(sizeof *csp, "allocating cuckoo header");
	csp->cu_bytes = sizeof *csp;
	csp->cu_numparts = minnt;
	csp->cu_maxparts = maxnt;
	csp->cu_numadd = 0;
	csp->cu_numdel = 0;
	csp->cu_numrelo = 0;
	csp->cu_find = 0;
	csp->cu_probe = 0;
	csp->cu_movecpt = 0;
	csp->cu_revd = *revdp;
	if (minsz < csp->cu_mintab) {
		fprintf(stderr,
	"Cuckoo: specified compartment size %zd is less than minimum size %zd\n", 
			minsz, csp->cu_mintab);
		fprintf(stderr,
	" ... will assume you know what you're doing.\n");
	}
	primsz = minsz - 1;
	for (ix = 0; ix < csp->cu_numparts; ix++) {
		primsz = nxtprime(primsz);
		alloctab(csp, ix, primsz);
	}
	for (; ix < cuck_MAXPARTS + 1; ix++) {
		csp->cu_ddd[ix] = 0;
		csp->cu_nent[ix] = 0;
	}
	return csp;
}

struct htab_tab *cuck_mustcreate(size_t minsz, size_t minnt, size_t maxnt,
		struct cuck_revdesc *revdp)
{
	struct htab_tab *csp = cuck_create(minsz, minnt, maxnt, revdp);
	if (csp)
		return csp;
	fprintf(stderr, "Failed to create cuckoo table: Aborting\n");
	exit(EXIT_MEMORY);
}

void cuck_free(struct htab_tab *csp)
{
	Uint ix;
	Bgroup *dp;

	for (ix = 0; ix < cuck_MAXPARTS + 1; ix++) {
		if (dp = csp->cu_ddd[ix])
			free(dp);
	}
	free(csp);
}

int htab_getnxt(struct htab_tab *csp, struct htab_trav *wp, Key *gp)
{
	int	retc;

	if (wp->wkreset)
		wp->wktab = 0;
	while ((retc = c_nextg(csp, wp, gp)) == 0) {
		/* an alternate test next, perhaps more reasonable,
		 * would be csp->cu_ddd[++(wp->wktab)] == NULL
		 *  Obviously(! :-)  they give the same result(sic).
		 * ... since we're not called while _movecpt() is running.
		 */
		if (++(wp->wktab) >= csp->cu_numparts)
			break;
		wp->wkreset = 1;
	}
	return retc;
}

/*
 * Develop a reversible scrambler,
 *	where n is max input,
 *	qq numb bits for quotient
 */
static void revgen(struct cuck_revdesc *rp, Key n, int qq, int paysz)
{
	Uint32	newn, a, res;
	Key	q;
	int	igbits = 0;

	/*
	 * Remember to check for bizarre params that
	 * would throw us for a loop.
	 */
	q = ((Key)1 << qq) - 1;
	for (igbits = 0; n > 255 << 22; n >>= 1, igbits++)
		;
	newn = n;
	rp->re_igb = igbits;
	rp->re_paysz = paysz;
	rp->re_bentry = paysz + qq;
	newn = nxtprime(newn - 1);
	a = newn * .61803398875 - 2;
	while ((res = mk_revh(a, newn)) == 0)
		++a;
	rp->re_im = res;
	rp->re_fm = a;
	rp->re_prim = newn;
	rp->re_mintab = ((long long)newn << igbits) / q;
	if (rp->re_mintab < 3)
		rp->re_mintab = 3;
#if	0
	printf(" q = %lld mintab = %ld newn = %ld\n",
			q, rp->re_mintab, newn);
#endif
}

void htab_create(struct htab_tab **p, int numtabs, struct htab_params *parp)
{
	struct cuck_revdesc revd;

	revgen(&revd, parp->maxkey, parp->qbits, parp->pbits);
	while (numtabs--)
		*p++ = cuck_mustcreate(revd.re_mintab,
				parp->mincuck, parp->maxcuck, &revd);
}
