
/* Copyright (c) 2008 James Dow Allen.  All Rights Reserved; see License. */
/*
 * Solve Peg Solitaire.
 * Program is straightforward and is mainly to
 *  demonstrate hash table method.
 * Because starting and goal positions of Peg Solitaire
 *   are identical, but color-reversed; and since forward
 *   and backward moves are identical, but color-reversed;
 *   we need only search halfway, handle a color-reversal
 *   seam, and then go backwards through already-found
 *   (but color-reversed) positions.
 */
#include	"js_simple.h"

typedef unsigned long long Key;
#include	"genhash.h"

typedef unsigned long long Grid;	/* same as Key ! */
#define	G1		((Grid)1)
#define	REVGRID(g)	((g) ^ (G1 << 33) - 1)

/* We need *two* tables, for odd and even positions,
 * so we can walk through one while writing the other.
 */
struct htab_tab *Tabp[2];

/*
 * Bit assignments for board:
 *	 .  .  .  0  1  2  .  .  .
 *	 .  .  .  3  4  5  .  .  .
 *	 .  6  7  8  9 10  1  2  .
 *	 .  3  4  5  6  7  8  9  .
 *	 . 20  1  2  3  4  5  6  .
 *	 .  .  .  7  8  9  .  .  .
 *	 .  .  . 30  1  2  .  .  .
 */

/* This table depicts all possible moves */
struct cposs {
	short a, b, c;
} Cposs[] = {
	30, 31, 32,
	24, 29, 32,
	0, 1, 2,	3, 4, 5,	6, 7, 8,	7, 8, 9,
	8, 9, 10,	9, 10, 11,	10, 11, 12,	13, 14, 15,
	14, 15, 16,	15, 16, 17,	16, 17, 18,	17, 18, 19,
	20, 21, 22,	21, 22, 23,	22, 23, 24,	23, 24, 25,
	24, 25, 26,	27, 28, 29,
	6, 13, 20,	7, 14, 21,	0, 3, 8,	3, 8, 15,
	8, 15, 22,	15, 22, 27,	22, 27, 30,	1, 4, 9,
	4, 9, 16,	9, 16, 23,	16, 23, 28,	23, 28, 31,
	2, 5, 10,	5, 10, 17,	10, 17, 24,	17, 24, 29,
	11, 18, 25,	12, 19, 26,
};

/* An elegant shape we might make our intermediate goal */
Grid Goal17 = G1 << 32 | (unsigned)0xf8010dbf;

void sholine(unsigned char c, int mask)
{
	int	i;

	printf("      ");
	for (i = 0; i < 7; i++, mask >>= 1, c >>= 1)
		printf("%c ", mask & 1 ? c & 1 ? 'X' : '.' : ' ');
	printf("\n");
}

void showgrid(Grid g, char *msg)
{
	printf("%s\n", msg);
	sholine(g << 2, 0x1c);
	sholine(g >> 1, 0x1c);
	sholine(g >> 6, 0x7f);
	sholine(g >> 13, 0x7f);
	sholine(g >> 20, 0x7f);
	sholine(g >> 25, 0x1c);
	sholine(g >> 28, 0x1c);
	printf("\n");
}

Grid mover(Grid g, struct cposs *p)
{
	return g ^ (G1 << p->a | G1 << p->b | G1 << p->c);
}

/*
 * determine if move *p is permitted in position g.
 * sens is 1/0 for reverse/forward moves.
 * permhv &1 allows horiz moves, permit &2 allows vertical moves
 */
int mable(int sens, Grid g, struct cposs *p, int permhv)
{
	if (!(permhv & (p->b - p->a == 1 ? 1 : 2)))
		return 0;
	return (sens == !(g & G1 << p->b))
		&& (g & G1 << p->a ?
			!(g & G1 << p->c) : g & G1 << p->c);
}

/* This application walks only one table at a time */
struct htab_trav Walkstate;

/* Walk through all positions with 'numpeg' pegs,
 * constructing the derived positions with 'numpeg+1' pegs.
 */
long genmoves(int numpeg)
{
	long	numpos = 0;
	long	posretr = 0;
	struct cposs *p;
	Grid	g, go;
	size_t	jj;
	int	retc;

	Walkstate.wkreset = -1;
	while (retc = htab_getnxt(Tabp[numpeg & 1], &Walkstate, &g)) {
		if (js_bitcnt(g) == numpeg) {
			++posretr;
			p = Cposs;
			for (jj = 0; jj < sizeof Cposs / sizeof *Cposs;
						jj++, p++) {
				if (mable(1, g, p, numpeg & 1 ? 1 : 2)) {
					go = mover(g, p);
					retc = htab_find(Tabp[!(numpeg & 1)],
							go, 0, HAmak, 0);
					if (retc >= 0)
						++numpos;
				}
			}
		}
	}
	printf("Retrieved %ld pos\n", posretr);
	return numpos;
}

Grid Mpath[33];

/* Find any move that will take us towards goal */
Grid looklin(int numpeg, Grid g, Grid revo)
{
	struct cposs *p;
	Grid	go;
	size_t	jj;

	p = Cposs;
	for (jj = 0; jj < sizeof Cposs / sizeof *Cposs; jj++, p++)
		if (mable(0, g, p, numpeg & 1 ? 1 : 2)) {
			go = mover(g, p);
			if (htab_find(Tabp[numpeg & 1], go ^ revo, 0, HLook, 0)) {
				return go;
			}
		}
	fprintf(stderr, "No move found by looklin, numpeg = %d\n", numpeg);
	showgrid(g, "Looklin input");
	exit(EXIT_SFAIL);
}

/* Halfway through, we need to hook forward to backward */
Grid good17(int numpeg)
{
	struct cposs *p;
	Grid	g, go;
	size_t	jj;
	int	retc;

	Walkstate.wkreset = -1;
	while (retc = htab_getnxt(Tabp[numpeg & 1], &Walkstate, &g)) {
		if (js_bitcnt(g) == numpeg) {
			p = Cposs;
			for (jj = 0; jj < sizeof Cposs / sizeof *Cposs; jj++, p++)
				if (mable(1, g, p, numpeg & 1 ? 1 : 2)) {
					go = mover(g, p);
					if (htab_find(Tabp[numpeg & 1],
							REVGRID(go), 0,
							HLook, 0)) {
						return go;
					}
				}
		}
	}
	fprintf(stderr, "Good17 failed\n");
	exit(EXIT_SFAIL);
}

/*
 * The table parameters are set (via makefile) by
 *  genparms for Tight Hash, but right here for Cuckoo.
 */
void maketables()
{
#ifdef	CUCKOO
	struct htab_params params;
	params.maxkey = (G1 << 33) - 1;
	params.qbits = 11;
	params.pbits = 0;
	params.mincuck = 2;
	params.maxcuck = 6;
	
	htab_create(Tabp, 2, &params);
#else
	htab_create(Tabp, 2, (struct htab_params *)0);
#endif
}

int main(int argc, char **argv)
{
	Grid	g;
	int	numpeg;
	long	numpos;
	char	posname[100];

	if (sizeof(Grid) != sizeof(Key)) {
		fprintf(stderr, "Grid and Key must be same!\n");
		fprintf(stderr, "This error can't happen :-)\n");
		exit(EXIT_SWCONF);
	}
	setlinebuf(stdout);
	maketables();
	/* Starting (and goal) position */
	g = G1 << 16;
	numpeg = 1;
	htab_find(Tabp[numpeg & 1], g, 0, HAmak, 0);
	Mpath[1] = g;
	Mpath[32] = REVGRID(g);

	for (numpeg++; numpeg < 17; numpeg++) {
		numpos = genmoves(numpeg - 1); 
		printf("With %d pegs, %ld positions\n", numpeg, numpos);
	}

	for (numpeg = 16; numpeg > 1; numpeg--) {
		if (numpeg == 16) {
			Mpath[17] = good17(32 - numpeg);
		} else {
			Mpath[33 - numpeg] = REVGRID(looklin(32 - numpeg,
					REVGRID(Mpath[32 - numpeg]),
					((Grid)0))); 
		}
		Mpath[numpeg] = looklin(numpeg, Mpath[numpeg + 1],
					(Grid)0); 
	}
	for (numpeg = 32; numpeg > 0; numpeg--) {
		sprintf(posname, "With %d pegs :", numpeg);
		showgrid(Mpath[numpeg], posname);
	}
	htab_sstat(Tabp[0]);
	htab_sstat(Tabp[1]);
	exit(EXIT_GOOD);
}
