
#define	GDEBUG	0

/* Copyright (c) 2008 James Dow Allen.  All Rights Reserved; see License. */
/*
 * Read an input file, and prepare counts of
 *   its words.
 * Purpose of program is to demonstrate
 *   Hash-based Digital search tree which is
 *   *very* memory efficient.
 * I won't provide a library-like interface until I get
 *   good input on what's needed.  For now, here's a simple
 *   comparison with straightforward tree:
 *
 * Straightforward tree goes down a level via:
 *	childp = thisp->link[k];
 *	if (childp == 0)
 *		childp = thisp->link[k] = allocate();
 *	thisp = childp;
 * Hash tree goes down a level via:
 *	htab_find(tp, TAG(thisp, k), 0, HHmak, &thisp);
 *
 * Straightforward tree increments leaf count, creating if necessary via
 *	leafp = thisp->link[k];
 *	if (leafp == 0)
 *		leafp = thisp->link[k] = allocate();
 *	leafp->count += addend;
 * Hash tree increments leaf count, creating if necessary via
 *	htab_find(tp, TAG(thisp, k), addend, HCadd, 0);
 *
 * This program allows 28 different symbols in
 *   a word, though this should be easy to change.
 *   The 28 symbols are a-z (with A-Z mapped to a-z]
 *   and ' and -
 */
#include	"js_simple.h"

typedef unsigned long Key;
#include	"genhash.h"

struct htab_tab *Tabp;
int	Numnode;

#define	NUM_SYMS	28
#define	LOG_NUMS	5

#define	MAX_WLENGTH	500

char	Validletts[]
	= "abcdefghijklmnopqrstuvwxyz'-ABCDEFGHIJKLMNOPQRSTUVWXYZ";

/* Convert a character to its integer code */
int let2ix(char c)
{
	char	*s;
	int	ix;

	if (c == '\0')
		return -2;
	s = strchr(Validletts, c);
	if (!s)
		return -1;
	ix = s - Validletts;
	/* Next switches upper to lower case */
	if (ix >= NUM_SYMS)
		ix -= NUM_SYMS;
	return ix;
}

#define	TAG(parent, k)	(((parent) << LOG_NUMS) + (k))

/* Add 1 to the count of the word in *lets,
 * creating any nodes needed along the way.
 */
int addword(int *lets, int cnt)
{
	int	par_tag, findres;

	par_tag = 0;
	while (--cnt) {
		if (GDEBUG) fprintf(stderr, "par_tag *lets = %d %d\n",
			par_tag, *lets);
		findres = htab_find(Tabp, TAG(par_tag, *lets), 0, HHmak, &par_tag);
		if (findres >= 0)
			++Numnode;
		++lets;
	}
	findres = htab_find(Tabp, TAG(par_tag, *lets), 1, HCadd, 0);
	if (findres >= 0)
		++Numnode;
	else
		findres = htab_AENTRY(findres);
	return findres;
}

void lookword(char *word)
{
	int	par_tag, findres = 0, tlet;
	char	*ow = word;

	par_tag = 0;
	while (tlet = *word++) {
		findres = htab_find(Tabp, TAG(par_tag, tlet - 'a'),
				0, HLook, &par_tag);
		fprintf(stderr, "Tag = %d\n", par_tag);
		if (findres == 0) {
			fprintf(stderr, "Look looked for *%s*, unfound resid = *%s*\n",
					ow, word - 1);
			return;
		}
	}
	fprintf(stderr, "Look looked for *%s*, got %x\n", ow, findres);
	return;
}

/* Process a line, finding words, incrementing their counts */
void handleline(char *s)
{
	int	ix, ret;
	int	letarr[MAX_WLENGTH], numlet;

	numlet = 0;
	while (1) {
		ix = let2ix(*s++);
		if (ix < 0) {
			if (numlet) {
			killword:
				ret = addword(letarr, numlet);
		if (GDEBUG) {
			fprintf(stderr, "Added *%.*s* cnt = %d\n",
				numlet, s - 1 - numlet,
				1 + htab_AENTRY(ret));
			lookword("lord");
			lookword("goddd");
		}
				numlet = 0;
			}
			if (ix == -2)
				return;
		} else {
			letarr[numlet++] = ix;
		}
		if (numlet > 40)
			goto killword;
	}
}

/* Display the counts */
void showwdcounts(char *sw, int swlen, int tag)
{
	int	ntag, i, findres;
	char	mbuff[MAX_WLENGTH];

	strcpy(mbuff, sw);
	mbuff[swlen + 1] = '\0';
	for (i = 0; i < NUM_SYMS; i++) {
		findres = htab_find(Tabp, TAG(tag, i), 0, HLook, &ntag);
		if (findres == 0)
			continue;
		findres = htab_AENTRY(findres);
		mbuff[swlen] = Validletts[i];
		if (findres)
			printf(" %7d occurrences of %s\n",
				findres, mbuff);
		showwdcounts(mbuff, swlen + 1, ntag);
	}
}

/* *I* use line-feeds in *my* files, and the buffer-size
 * following is **plenty** big enough for *me* !
 */
#define	MAXLINE	15000000
char	Wbuf[MAXLINE];

int main(int argc, char **argv)
{
	Uint64	maxkey;

	setlinebuf(stdout);
	if (1 << LOG_NUMS < NUM_SYMS) {
		printf("Oops!  2^Log_nums must be at least Num_syms\n");
		exit(EXIT_SWCONF);
	}
	htab_create(&Tabp, 1, (struct htab_params *)0);
	if (htab_maxpoint(Tabp) < 1 << LOG_NUMS) {
		printf("Oops!  Qsize too small\n");
		exit(EXIT_SWCONF);
	}
	maxkey = htab_maxtag(Tabp);
	maxkey = TAG(maxkey, NUM_SYMS - 1);
	if (maxkey > htab_maxkey(Tabp)) {
		printf("Oops!  Key range too small (%llu > %llu)\n",
			maxkey, htab_maxkey(Tabp));
		exit(EXIT_SWCONF);
	}
	while (fgets(Wbuf, MAXLINE, stdin)) {
		handleline(Wbuf);
	}
	printf("Created %d nodes.\n", Numnode);
	showwdcounts("", 0, 0);
	htab_sstat(Tabp);
	exit(EXIT_GOOD);
}
