
/*
 * j_random.c  --  Random number generation and routines
 * Copyright (c) 2013-2020 by James Dow Allen
 *
 *  This program can be distributed or modified subject to
 *  the terms of the GNU General Public License (version 2
 *  or, at your option, any later version) as published by
 *  the Free Software Foundation.  All Wrongs Reversed.
 *
 * The functions form Layers.
 *	Layer 1) The underlying generators
 *	Layer 2) rand32() - trivial interface
 *	Layer 3) rand_bits(), rand_vprob, assisted by rand_stow, rkeep, rkcnt
 *	Layer 4) rand_index(), rand_decide(), prime_ir()
 *	Layer 5) misc. functions
 *		which call rand_index, rand_decide or rand_vprob
 *	(Hmmm. I hope this isn't sounding worse than it is!  :-) )
 * 
 * In addition to the 18 functions documented in j_random.h
 * there are 7 static functions.
 * 	ma_random, me_random, ma_refill
 *		- two excellent generators and a subroutine
 *	rand_stow, prime_ir
 *		- manipulate left-over entropy
 *	locbit	- similar to log2()
 *	rand32	- internal entry to get 32 fresh bits
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<math.h>
#define		J_PI	3.14159265358979323846
#include	"j_random.h"

#ifndef		BITS_FOR_P
#define		BITS_FOR_P	31
#endif

int	wordsused = 0;

/* ******************* Layer 1 ***************** */
/* The PRNGs themselves */

/* ************ Marsaglia KISS adapted from public domain ************ */
#define	CSIZ	41790

static uint32_t Q[CSIZ],
	seed1	= 6,
	seed	= 6,
	indx	= CSIZ + 1,
	carry	= 362436,
	xcng	= 1236789,
	xs	= 521288629;

static void ma_refill(void)
{
	uint32_t i;
	uint64_t t;

	if (indx > CSIZ)
	for (indx = 0; indx < CSIZ; indx++) {
		xs ^= xs << 13;
		xs ^= (uint32_t)xs >> 17;
		xs ^= xs >> 5;
		xcng = 69609 * xcng + 123;
		Q[indx] = xcng + xs;
	}
	for(i = seed1 % CSIZ; i < CSIZ; i++) {
		t = (long long)7010176 * Q[i] + carry;
		carry = t >> 32;
		Q[i] = ~t;
	}
	for(i = 0; i < seed1 % CSIZ; i++) {
		t = (long long)7010176 * Q[i] + carry;
		carry = t >> 32;
		Q[i] = ~t;
	}
	seed1 = indx = 0;
}

static uint32_t ma_random(void)
{
	uint32_t	resu;
	if (indx >= CSIZ)
		ma_refill();
	xs ^= xs << 13;
	xs ^= (uint32_t)xs >> 17;
	xs ^= xs >> 5;
	xcng = 69609 * xcng + 123;
	resu = Q[indx++] + xcng + xs;
	seed *= 1103515245;
	seed += 12345;
	resu ^= seed;
	seed *= 1103515245;
	seed += 12345;
	resu ^= seed >> 16;
	return resu;
}

/* ************ Mersenne Twister adapted from public domain ************ */
#define	MTsize	624
static uint32_t MT[MTsize + 1];
static int MTindex = MTsize + 2;

static uint32_t me_random(void)
{
	uint32_t	y, y1, *p;
	int	i;

	if (MTindex >= MTsize) {
		if (MTindex >= MTsize + 2) {
			MT[0] = seed | 0x4000000;
			MT[1] = seed ^ 0x5555555;
			for (i = 2; i < MTsize; i += 2) {
				MT[i] = 1812433253l
					* (MT[i-2] ^ MT[i-2] >> 30) + 1;
				MT[i+1] = 69069 * MT[i-1];
			}
		}
		y1 = MT[0];
		for (p = MT; p < MT + 227; p++) {
			y = y1 & 0x80000000;
			y1 = p[1];
			y += y1 & 0x7fffffff;
			p[0] = (p[397] ^ y >> 1) ^ (y & 1 ? 0x9908b0df : 0);
		}
		MT[MTsize] = MT[0];
		for (; p < MT + MTsize; p++) {
			y = y1 & 0x80000000;
			y1 = p[1];
			y += y1 & 0x7fffffff;
			p[0] = (p[-227] ^ y >> 1) ^ (y & 1 ? 0x9908b0df : 0);
		}
		MTindex = 0;
	}
	y = MT[MTindex++];
	y ^= y >> 11;
	y ^= y << 7 & 0x9d2c5680;
	y ^= y << 15 & 0xefc60000;
	y ^= y >> 18;
	return y;
}

/* ************ Some simple PRNGs we don't use ************ */
/* Bob Jenkins' PRNG */
struct ranctx {
	uint32_t a, b, c, d;
} burtle_struc;

#define rot(x,k) (((x)<<(k))|((x)>>(32-(k))))
uint32_t burtle_rand(/* struct ranctx *x */)
{
	uint32_t e;
	struct ranctx *x = &burtle_struc;

	e = x->a - rot(x->b, 27);
	x->a = x->b ^ rot(x->c, 17);
	x->b = x->c + x->d;
	x->c = x->d + e;
	x->d = e + x->a;
	return x->d;
}

void burtle_seed(/* struct ranctx *x, */ uint32_t seed)
{
	int i;
	struct ranctx *x = &burtle_struc;

	x->a = 0xf1ea5eed, x->b = x->c = x->d = seed;
	for (i=0; i<20; ++i) {
		(void)burtle_rand(/* x */);
	}
}

/*
 *  Here is a good generator, somewhat faster
 * than the above, though not linked to by our library
 */
static uint64_t fastr_a, fastr_b, fastr_c, fastr_counter, fastr_save;

uint32_t fast_random(void)
{
	uint32_t tmp;

	if (fastr_save) {
		tmp = fastr_save >> 32;
		fastr_save = 0;
		return tmp;
	}
	fastr_save = fastr_a + fastr_b + fastr_counter++;
	fastr_a = fastr_b ^ (fastr_b >> 12);
	fastr_b = fastr_c + (fastr_c << 3);
	fastr_c = ((fastr_c << 25) | (fastr_c >> 39)) + fastr_save;
	return fastr_save;
}

void fast_rseed(uint64_t seed)
{
	int	i;

	fastr_a = fastr_b = fastr_c = seed | ((uint64_t)seed) << 32;
	fastr_counter = 1;
	fastr_save = 0;
	for (i = 0; i < 10; i++)
		fast_random();
}

void rand_seed(unsigned int x)
{
	int	i;

	burtle_seed(seed);
	fast_rseed(x | ((uint64_t)x) << 32);
#ifndef	JR_FASTER
	seed = seed1 = x;
	indx = CSIZ + 1;
	MTindex = MTsize + 2;
	for (i = 99 + ((seed ^ seed << 5) & 0x3ff); i; i--)
		me_random() & 0x882 ? me_random() : ma_random();
#endif
}

/* ******************* Layer 2 ***************** */
/* The PRNGs in Layer 1 are invoked only by rand32()
 * Layers 3-5 get all their random bits directly or
 *	ondirectly from rand32().
 *
 * We make rand32() 'static' -- (change it if you will) -- just
 *	to avoid polluting the namespace in user's head.
 *	rand_bits(32) will call rand32() with modest additional
 *	overhead.  (And user calls SHOULD be going to rand_index,
 *	rand_decided, rand_{,v}prob which obviate even the modest
 *	overhead.
 */
static uint32_t rand32(void)
{
	wordsused += 1;
	uint32_t rr;
#ifdef	JR_FASTER
	return fast_random();
#else
/* Our PRNG is the sum of two excellent and independent PRNGs */
	rr = me_random();
	return rr + ma_random();
#endif
}

/* ******************* Layer 3 ***************** */
/* When we want less than 32 bits, allow the excess bits
 *	to be stowed for future use.
 * Routines in Layers 4 and 5 will call rand_bits()
 *  (rand32() can be substituted for its synonym rand_bits(32).)
 */

/* We store rkcnt number of random bits in rkeep */
static uint64_t	rkeep = 0;
static uint32_t	rkcnt = 0;

uint32_t rand_bits(unsigned int cnt)
{
	uint32_t	x;

	if (cnt > 32) {
		fprintf(stderr, "Random words used = %d\n", wordsused);
		return 0;
	}
	switch (cnt) {
	case 32:
		return rand32();
	case 31:
		return rand32() >> 1;
	case 30:
		return rand32() >> 2;
	default:
		if (rkcnt < cnt) {
			x = rand32();
			rkeep = rkeep << (32 - cnt) | x >> cnt;
			rkcnt += 32 - cnt;
		} else {
			rkcnt -= cnt;
			x = rkeep;
			rkeep >>= cnt;
		}
		return x & ((uint32_t)1 << cnt) - 1;
	}
}

/* ******************* Layer 4 ***************** */
/* Most callers don't want a fixed number of bits.
 * They want either:
 *	a probability -- call rand_prob() or rand_vprob(prec),
 *	a small uniform integer -- call rand_index(range), or
 *	a binary decision -- call rand_decide(p).
 * rand_index and rand_decide know how to save "unused entropy"
 *	for later use.  They do this using prime_ir, rand_stow().
 */

/* Return a uniform random variable from 0 to 1 */
double rand_prob(void)
{
	return (rand32() | 1) / (4.0 * (1ul << 30));
}

/* Return a uniform random variable, using only prec random bits */
double rand_vprob(unsigned int prec)
{
	if (prec > 30)
		return (rand32()     | 1) / (4.0 * (1ul << 30));
	return (rand_bits(prec) << 1 | 1) / (4.0 * (1ul << prec-1));
}

/* Ineptitude! By returning bits via rkeep instead of
 * into ir_vari we may get wasteful!
 * Turn on the fprintf to see if this ever happens.
 */
static void rand_stow(uint32_t x, int cnt)
{
	if (rkcnt + cnt > 64) {
		if (0)
		fprintf(stderr, "Discarding %d bits in rand_stow()\n",
			rkcnt + cnt - 64);
		cnt = 64 - rkcnt;
	}
	if (cnt < 32)
		x &= ((uint32_t)1 << cnt) - 1;
	rkeep = rkeep << cnt | x;
	rkcnt += cnt;
}

static char Lbitp[] = {
#define	LB_FILL(x)	x, 1,  2, 1, 3, 1, 2, 1
#define	LB_FIL2(x)	LB_FILL(x), LB_FILL(4), LB_FILL(5), LB_FILL(4)

	LB_FIL2(0), LB_FIL2(6), LB_FIL2(7), LB_FIL2(6),
	LB_FIL2(8), LB_FIL2(6), LB_FIL2(7), LB_FIL2(6),
};

static int locbit(uint32_t range)
{
	int	a;

	a = Lbitp[range & 0xff];
	if (a) return a;
	range >>= 8;
	a = Lbitp[range & 0xff];
	if (a) return a + 8;
	range >>= 8;
	a = Lbitp[range & 0xff];
	if (a) return a + 16;
	range >>= 8;
	a = Lbitp[range & 0xff];
	if (a) return a + 24;
	return 0;
}

/* Maintain ir_vari as a random integer on {0, 1, ..., ir_range-1} */
static uint32_t ir_range = 1, ir_vari = 0;

static void prime_ir()
{
	int	t = 0;

	if (ir_range < 1 << 11) {
		ir_range <<= 21;
		ir_vari <<= 21;
		ir_vari |= rand_bits(21);
	} else if (ir_range < 1 << 21) {
		ir_range <<= 11;
		ir_vari <<= 11;
		ir_vari |= rand_bits(11);
	}
	if (ir_range < 1 << 26) {
		ir_range <<= 5;
		ir_vari <<= 5;
		ir_vari |= rand_bits(5);
	}
}

unsigned int simp_index(int range)
{
	unsigned int x;

	do {
		x = rand32();
	} while (x + range < x);
	return x % range;
}

unsigned int rand_index(int range)
{
	uint32_t	x, y, maxy;

	if (range < 2)
		return 0;
	if ((range & range - 1) == 0)
		return rand_bits(locbit(range) - 1);
	if (range > 1 << 21) {
		do {
			x = rand32();
			y = x / range;
		} while (y * (range + 1) < y);
                return x % range;
	}
	while (1) {
		prime_ir();
		maxy = ir_range / range;
		y = ir_vari / range;
		if (y < maxy)
			break;
		ir_range %= range;
		ir_vari %= range;
	}
	x = ir_vari % range;
	ir_vari = y;
	ir_range = maxy;
	return x;
}

/* Return 1 with probability p */
int rand_decide(double p)
{
	uint32_t	x;
	int	j;

	if (p < 0.000001) {
		if (p <= 0)
			return 0;
		j = rand_index(65537);
		if (j) {
			rand_stow(j - 1, 16);
			return 0;
		}
		p *= 65537;
	}
	if (p < 0.0004) {
		j = rand_index(2049);
		if (j) {
			rand_stow(j - 1, 11);
			return 0;
		}
		p *= 2049;
	} else if (p > 0.99) {
		return ! rand_decide(1 - p);
	}
	prime_ir();
	x = ir_range * p + 0.5;
	if (ir_vari < x) {
		ir_range = x;
		return 1;
	} else {
		ir_vari -= x;
		ir_range -= x;
		return 0;
	}
	/*
	 * Statistical parameters for binomial distribution with p, N:
	 *  (where q = 1-p)
	 * Mean:	N*p
	 * Std-dev:	sqrt(N*p*q)
	 * Skew:	(1-p) / sqrt(N*p*q)
	 * Kurtosis:	3 + (1-6*p*q) / (N*p*q)
	 */
}

/* ******************* Layer 5 ***************** */
/* Higher-level functions, mostly calling
 *	rand_index(), rand_prob() and rand_decide().
 */

/* Set a random k-sized combo of {0,1,...,(m-1)}, returning 1 if sucessful */
int rand_combo(int result[], int m, int k)
{
	double	p;

	if (k < 0 || m < k)
		return 0;
	result += k;
	while (k) {
		if (m == k || rand_decide(k / (double)m)) {
	/* predecrement avoids setting an (unused) possibly
	 * invalid pointer.  Spotted by Ben Bacarisse.
	 */
			*--result = m - 1;
			k--;
		}
		m--;
	}
	return 1;
}

/* Shuffle the m-sized array[] */
void rand_shuffle(int array[], int m)
{
	int	t, j;

	for (; m > 1; m--, array++) {
		j = rand_index(m);
		t = array[j];
		array[j] = array[0];
		array[0] = t;
	}
}

/* Shuffle the m-sized array[] */
void rand_shshuff(short array[], int m)
{
	int	t, j;

	for (; m > 1; m--, array++) {
		j = rand_index(m);
		t = array[j];
		array[j] = array[0];
		array[0] = t;
	}
}

/* Shuffle the m-sized array[] */
void rand_llshuff(long long array[], int m)
{
	int	j;
	long long t;

	for (; m > 1; m--, array++) {
		j = rand_index(m);
		t = array[j];
		array[j] = array[0];
		array[0] = t;
	}
}

/* Return a uniform random variable from `minv' to `maxv' */
double rand_unif(double minv, double maxv)
{
	return minv + (maxv - minv) * rand_vprob(BITS_FOR_P);
	/*
	 * With R = maxv - minv
	 * Mean:	minv + 0.5*R
	 * Std-dev:	0.2887 * R
	 * Skew:	0
	 * Curtosis:	1.8
	 */
}

/* Return a geometric random variable on (1,2,3,...) with parameter prob */
int rand_geomd(double prob)
{
	double	t;

	t = log(rand_vprob(BITS_FOR_P));
	t /= log(1 - prob);
	t += 1;
	return t;
	/*
	 * Mean:	1/p
	 * Std Dev:	sqrt(1-p) / p
	 * Skew:	(2-p) / sqrt(1-p)
	 * Kurtosis:	9 + p*p/(1-p)
	 */
}

/* Return an exponential random variable with parameter lambda */
double rand_expon(double lambda)
{
	return -log(rand_vprob(BITS_FOR_P)) / lambda;
	/*
	 * Mean:	lambda
	 * Std-dev:	lambda
	 * Skew:	2
	 * Kurtosis:	9
	 */
}

/* Return a Weibull random variable with parameters lambda, k */
double rand_weibull(double lambda, double k)
{
	double	t;

	t = -log(rand_vprob(BITS_FOR_P));
	return exp(log(t) / k) / lambda;
}

/* Return a Poisson random variable with mean and variance = ln(m) */
int rand_poisson(double m)
{
	static double z = 0.5;
	int	n;

	z *= m;
	for (n = 0; z >= 1.0; n += 1)
		z *= rand_vprob(BITS_FOR_P);
	return n;
	/*
	 * Mean:	m
	 * Std-dev:	sqrt(m)
	 * Skew: 	1 / sqrt(m)
	 * Kurtosis:	3 + 1/m
	 */
}

/* Return a gaussian random variable with mean `meanv' and std-dev `sdevv' */
double rand_gaussian(double meanv, double sdevv)
{
	static int iset = 1;
	static double	v1, v2;

	if (iset ^= 1)
		return meanv + sdevv * v2 * sin(v1);
	v1 = rand_vprob(BITS_FOR_P) * 2 * J_PI;
	v2 = sqrt(-2 * log(rand_vprob(BITS_FOR_P)));
	return meanv + sdevv * v2 * cos(v1);
	/*
	 * Mean:	meanv
	 * Std-dev:	sdevv
	 * Skew:	0
	 * Kurtosis:	3
	 */
}

/* Return a cauchy random variable */
double rand_cauchy(void)
{
	static int beenhere = 0;
	static double	y2;
	double	y1, v1, v2;

	if (! beenhere) {
		v1 = rand_vprob(BITS_FOR_P) * 2 * J_PI;
		v2 = sqrt(-2 * log(rand_vprob(BITS_FOR_P)));
		y2 = v2 * cos(v1);
		beenhere = 1;
	}
	v1 = rand_vprob(BITS_FOR_P) * 2 * J_PI;
	v2 = sqrt(-2 * log(rand_vprob(BITS_FOR_P)));
	y1 = v2 * sin(v1);
	y1 /= y2;
	y2 = v2 * cos(v1);
	return y1;
}

/* Return a random point on the surface of a N-dim hyperball */
void rand_surfpoint(double coord[], double radius, int n)
{
	int	i;
	double	dsq, t;

	for (dsq = i = 0; i < n; i++) {
		t = coord[i] = rand_gaussian(0.0, 10.0);
		dsq += t * t;
	}
	dsq = radius / sqrt(dsq);
	for (i = 0; i < n; i++)
		coord[i] *= dsq;
}

/* Return a random point in interior of a N-dim hyperball */
double rand_inball(double coord[], double radius, int n)
{
	int	i;
	double	dadj;

	rand_surfpoint(coord, 1.0, n);
	dadj = radius * pow(rand_vprob(BITS_FOR_P), 1.0 / n);
	for (i = 0; i < n; i++)
		coord[i] *= dadj;
	return dadj;
}

/*
 * On entry to rand_setwalker(), wp[] is a histogram with
 *  x1 - value, x2 - unused, prob1 -- frequency of x1, arbitrarily scaled.
 * On exit from rand_setwalker(), wp[] has been modifed for use
 *  by rand_walker().
 *
 *	struct walkerent {
 *		int     x1, x2;
 *		double  prob1;
 *	} wp[size];
 */

/* Return value originally set in wp[]->x1 according to set distribution */
int rand_walker(struct walkerent *wp, unsigned int size)
{
	wp += rand_index(size);
	return rand_decide(wp->prob1) ? wp->x1 : wp->x2;
}

#define	TINY	0.0000000001
/* Given a histogram in wp[0 ... (size-1)], prepare for rand_walker() */
void rand_setwalker(struct walkerent *wp, int size)
{
	int	i, is, il;
	double	p;
	struct walkerent wt;

	for (p = i = 0; i < size; i++) {
		if (wp[i].prob1 < 0)
			wp[i].prob1 = 0;
		p += wp[i].prob1;
	}
	/* Next would be a user error, but let's not be dividing by zero! */
	if (p < TINY) {
		wp[0].prob1 += TINY - p;
		p = TINY;
	}
	p = size / p;
	for (i = 0; i < size; i++)
		wp[i].prob1 *= p;
	for (is = 0, il = size - 1; ; ) {
		while (wp[is].prob1 < 1)
			++is;
		while (wp[il].prob1 > 1)
			--il;
		if (is >= il)
			break;
		wt = wp[is];
		wp[is++] = wp[il];
		wp[il--] = wt;
	}
	for (i = 0; i < size; i++) {
		wp[i].x2 = wp[is].x1;
		if (i >= is)
			++is;
		else if (((wp[is].prob1 -= 1 - wp[i].prob1) < 1)
				&& is < size - 1)
			++is;
	}
}

/*
 * rand_ewalker, rand_esetwalker provide an interface identical to
 *	rand_walker, rand_setwalker except that parsimony
 *	of random bits is the goal, instead of speed.
 */
int rand_ewalker(struct walkerent *wp, unsigned int size)
{
	while (--size) {
		if (rand_decide(wp->prob1))
			break;
		wp++;
	}
	return wp->x1;
}

static int wecomp(const void *p1, const void *p2)
{
	double	f1 = ((struct walkerent *)p1)->prob1;
	double	f2 = ((struct walkerent *)p2)->prob1;
	return f1 < f2 ? 1 : f1 > f2 ? -1 : 0;
}

void rand_esetwalker(struct walkerent *wp, int size)
{
	double	resid, tt;
	int	i;

	for (tt = i = 0; i < size; i++)
		tt += wp[i].prob1;
	if (tt < TINY) {
		wp[0].prob1 = 2;
		return;
	}
	for (i = 0; i < size; i++)
		wp[i].prob1 /= tt;
	qsort(wp, size, sizeof *wp, wecomp);
	resid = 1;
	for (i = 0; i < size; i++) {
		tt = wp[i].prob1;
		wp[i].prob1 = resid < TINY ? 1 : tt / resid;
		resid -= tt;
	}
}
