
/*
 * j_random.h  --  Random number generation and routines
 * Copyright (c) 2013 by James Dow Allen
 *
 *  This program can be distributed or modified subject to
 *  the terms of the GNU General Public License (version 2
 *  or, at your option, any later version) as published by
 *  the Free Software Foundation.  All Wrongs Reversed.
 *
 * Here's a simple shortcut.  We don't use it and can't decide whether
 * to make it a function or a macro anyway, so here it is, shown as comment.
 *	Given p random uniform on [0,1), produce random uniform 0 <= k < E,
 *		leaving p independent and still random uniform on [0,1)
 	k = (int)(p *= E)
	if (k >= E) k = E-1; // paranoia
	p -= k;
 */

#define	CHEAP_RK(k, p, siz)		\
	p -= (k = p *= (siz))

#if	0
#define	JR_FASTER
#endif

/* Modify next to spend fewer random bits on simulations */
#define		BITS_FOR_P	31

/* Seed or reseed the random generator */
void rand_seed(unsigned int seed);

/* An alternate fast 32-bit generator */
uint32_t fast_random(void);

/* Fast generator accepts 64-bit seed */
void fast_rseed(uint64_t seed);

/* Return count random bits.  0 < count < 33 */
uint32_t rand_bits(unsigned int count);


/* Return x, random uniform 0 < x < 1
 *
 * One of 2 billion values is returned:
 * rand32() | 1     rand_prob()
 * --------       -----------------
 * 00000001 -->   0.000000000232831
 * 00000003 -->   0.000000000698492
 * 00000005 -->   0.000000001164153
 *	     ...
 * 7ffffffd -->   0.499999999301508
 * 7fffffff -->   0.499999999767169
 * 80000001 -->   0.500000000232831
 * 80000003 -->   0.500000000698492
 *	     ...
 * fffffffb -->   0.999999998835847
 * fffffffd -->   0.999999999301508
 * ffffffff -->   0.999999999767169
 */
double rand_prob(void);

/* Return random uniform 0 < x < 1 using only 1 < prec < 30 bits */
double rand_vprob(unsigned int prec);

/* Return 1 with probability prob */
int rand_decide(double prob);

/* Set a random k-sized combo of {0,1,...,(m-1)}, returning 1 if sucessful */
int rand_combo(int result[], int m, int k);

/* Shuffle the m-sized array[] */
void rand_shuffle(int array[], int m);
void rand_shshuff(short array[], int m);
void rand_llshuff(long long array[], int m);

/* Return random uniform 0 <= x < range */
unsigned int rand_index(int range);

/* Return a uniform random variable from `minv' to `maxv' */
double rand_unif(double minv, double maxv);

/* Return a geometric random variable (>= 1) with parameter prob */
int rand_geomd(double prob);

/* Return an exponential random variable with parameter lambda */
double rand_expon(double lambda);

/* Return a Weibull random variable with parameters lambda, k */
double rand_weibull(double lambda, double k);

/* Return a Poisson random variable with mean and variance = ln(e_to_p) */
int rand_poisson(double e_to_p);

/* Return a gaussian random variable with mean `meanv' and std-dev `sdevv' */
double rand_gaussian(double meanv, double sdevv);

/* Return a cauchy random variable */
double rand_cauchy(void);

/* Set a random point on the surface of a numdim-dimensional hyperball */
void rand_surfpoint(double coord[], double radius, int numdim);

/* Set a random point in interior of a numdim-dimensional hyperball */
double rand_inball(double coord[], double radius, int numdim);

struct walkerent {
	int	x1, x2;
	double	prob1;
};

/*
 * On entry to rand_setwalker(), wp[] is a histogram with
 *  x1 - value, x2 - unused, prob1 -- frequency of x1, arbitrarily scaled.
 * On exit from rand_setwalker(), wp[] has been modifed for use
 *  by rand_walker().
 */

/* Given a histogram in wp[0 ... (size-1)], prepare for rand_walker() */
void rand_setwalker(struct walkerent *wp, int size);

/* Return value originally set in wp[]->x1 according to set distribution */
int rand_walker(struct walkerent *wp, unsigned int size);

/* Compatible with rand_setwalker, rand_walker but minimizing
 *  consumption of random bits.
 */
void rand_esetwalker(struct walkerent *wp, int size);
int rand_ewalker(struct walkerent *wp, unsigned int size);
