
/* Merge two .srt subtitle files into a single
 *	bilingual .srt file. .sub files are also
 *	supported (output will still be .srt)
 *	but not well tested.
 * Optionally the timestamps can be subjected to an
 *	affine transform.  To do the transform on a
 *	single .srt file, use "none" as one of the file-names. 
 * The most difficult and confusing part of the program
 *	handles overlapping subtitles.  Sometimes one
 *	subtitle file will leave a 2-sentence subtitle intact,
 *	while the other will break it into separate
 *	1-sentence subtitles.  In this case, the longer subtitle
 *	must be duplicated and inserted into the output TWICE.
 * To merge THREE subtitle files together, merge two; then
 *	merge the output of that with the 3rd file.
 *
 * Coping with overlap in subtitles is ridiculously complicated and
 *	still doesn't work well.  Some code after 'goto skipkluge'
 *	was intended to help, but made matters worse in some cases.
 *
 * An interesting situation is a 3-sentence speech that gets split
 *	2&1 and 1&2 in the two files.  Our code doesn't deduce
 *	sentences; it uses subtitles, but the first one might be
 *	4&2 seconds, the latter 2&4 seconds.  Now what do we do?
 *	(a) Arrange it as 3 seconds and 3 seconds.
 *	(b) generate three subtitle records (2 seconds each):
 *		A1/A2/B1  ; A1/A2/B2/B3 ; A3/B2/B3.
 * (b) is easier to achieve and better BUT the A1/A2 subtitle will
 *	jump annoyingly on the screen of 'B1' and 'B2/B3' have
 *	an unequal number of lines.
 * (In some variations, we get little 'runt' subtitle records
 *	jut half a second long or so.)
 *
 * A simple solution is adopted: Count the number of lines in each
 *	B subtitle record: add one (or two) newlines
 *	between A and B text to get the count to 3.
 */

#define		_GNU_SOURCE
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#if	0
#define	VVBUGGY
#endif
#define	KLUGFACT	0.27

int	Iflag, Cflag, Onlyonefile, Supwrap;
int	Maxvvv = 1;

struct subt {
	int	tstart, tstop;
#ifdef	VVBUGGY
	int	ktstart, ktstop;
#endif
	int	subfno;
	int	numline;
	int	vertu;
	unsigned char	*stext[12];
} Subt[20000];

struct timadj {
	double mpy, add;
} Timadj[3] = {
	1, 0,
	1, 0,
	1, 0,
};

int subcomp(const void *p1, const void *p2)
{
	int	j1 = ((struct subt *)p1)->tstart;
	int	j2 = ((struct subt *)p2)->tstart;
	return j1 - j2;
}

int subcomp2(const void *p1, const void *p2)
{
	int	j1 = ((struct subt *)p1)->tstop;
	int	j2 = ((struct subt *)p2)->tstop;
	return j1 - j2;
}

int time2int(unsigned char *s, int fno, int notsrt)
{
	int	hr, mi, se, ms, cnt;

	if (notsrt) {
		cnt = sscanf(s, "%d:%d:%d.%d", &hr, &mi, &se, &ms);
		ms *= 10;
	} else {
		cnt = sscanf(s, "%d:%d:%d,%d", &hr, &mi, &se, &ms);
	}
	if (cnt != 4) {
		fprintf(stderr, "Count in time2int() %d *%s*\n", cnt, s);
		exit(1);
	}
	ms += ((hr * 60 + mi) * 60 + se) * 1000;
	ms = ms * Timadj[fno].mpy + Timadj[fno].add;
	return ms;
}

unsigned char *int2time(int tim)
{
	static unsigned char s[20][40];
	static int bcnt = 0;
	int	hr, mi, se, ms;

	bcnt = (bcnt + 1) % 20;
	ms = tim % 1000;
	se = tim / 1000;
	mi = se / 60;
	se -= mi * 60;
	hr = mi / 60;
	mi -= hr * 60;
	sprintf(s[bcnt], "%02d:%02d:%02d,%03d", hr, mi, se, ms);
	return s[bcnt];
}

unsigned char	Buff[10000];
FILE	*The_file;
int	Firstinf;
int	Boms;

unsigned char	*jgets(unsigned char *bf)
{
	unsigned char	*res;
	int	len;

	*bf = '\0';
	res = fgets(bf, 9999, The_file);
	if (! res)
		return res;
	len = strlen(bf) - 1;
	if (len >= 0 && (bf[len] == '\n' || bf[len] == '\r'))
		--len;
	if (len >= 0 && (bf[len] == '\n' || bf[len] == '\r'))
		--len;
	bf[len + 1] = '\0';
	if (Firstinf && res[0] == 0xef
			&& res[1] == 0xbb && res[2] == 0xbf) {
		res += 3;
		Boms += 1;
		if (0) fprintf(stderr, "Bom detected\n");
	}
	Firstinf = 0;
	return res;
}

unsigned char lastc(unsigned char *r)
{
	return r[strlen(r) - 1];
}

void clean_br(char *p)
{
	p[0] = ' ';
	p[1] = ' ';
	p[2] = '\n';
	p[3] = ' ';
}

void fixuptext(struct subt *p)
{
	int	nl = p->numline;
	int	i, lc, nomerge = 0;
	int	num_ladded = 0;
	unsigned char *brp;

	for (i = 0; i < nl; i++) {
		while (brp = strstr(p->stext[i], "[br]")) {
			clean_br(brp);
			++num_ladded;
		}
		while (brp = strstr(p->stext[i], "[BR]")) {
			clean_br(brp);
			++num_ladded;
		}
	}
	for (i = 0; i < nl - 1; i++) {
		lc = lastc(p->stext[i]);
		if (1
				&& Supwrap == 0
				&& nomerge == 0
				&& ':' != lastc(p->stext[i+1])
				&& '-' != p->stext[i+1][0]
				&& '.' != lc
				&& '?' != lc
				&& '!' != lc
				&& strlen(p->stext[i])
					+ strlen(p->stext[i+1])
					< (p->subfno == 1 ? 62 : 75)
		) {
			strcat(p->stext[i], " ");
			nomerge = 1;
			--num_ladded;
		} else {
			strcat(p->stext[i], "\n");
			nomerge = 0;
		}
	}
	p->vertu = p->numline + num_ladded;
}

char Ctext[100];
int	Cordia = 0;

void printsubs(struct subt *p)
{
	int	imarkit = Iflag == p->subfno;
	int	cmarkit = Cflag == p->subfno;
	int	nl = p->numline;
	int	i, lc, nomerge = 0;
	unsigned char *brp;

	if (Cordia && p->subfno == 2)
		printf("{\\fnCordiaUPC}");
	if (cmarkit)
		printf("%s", Ctext);
	if (imarkit)
		printf("<i>");
	for (i = 0; i < nl; i++) {
		printf("%s", p->stext[i]);
	}
	printf("%s%s\n",
			imarkit ? "</i>" : "",
			cmarkit ? "</font>" : "");
}

void printpair(struct subt *q1, struct subt *q2)
{
	if (q1->subfno == 1) {
		printsubs(q1);
		if (q2->vertu > Maxvvv) {
		} else if (q2->vertu == Maxvvv) {
			printf(" \n");
		} else {
			printf(" \n");
			printf(" \n");
		}
		printsubs(q2);
	} else {
		printsubs(q2);
		if (q1->vertu > Maxvvv) {
		} else if (q1->vertu == Maxvvv) {
			printf(" \n");
		} else {
			printf(" \n");
			printf(" \n");
		}
		printsubs(q1);
	}
}

int main(int argc, char **argv)
{
	unsigned char	*pstop, *bufp;
	int	cntr, fno, firstinf;
	struct subt *p, *q, *q1, *q2, *endp;
	int	tim1, tim0, tim3, timmed, ktim1;
	int	Notsrt;
	int	stopadj, startadj, adjcount, adjf;

	setlinebuf(stdout);
	Iflag = 0;
	Maxvvv = 1;
	while (1) {
		if (!strcmp(argv[1], "-Cordia")) {
			argc -= 1;
			argv += 1;
			Cordia = 1;
		}
		if (argc >= 3 && !strncmp(argv[1], "-C", 2)) {
			Cflag = argv[1][2] == '2' ? 2 : 1;
			sprintf(Ctext, "<font color=\"%s%s\">",
					argv[2][0] == '#' ? "" : "#",
					argv[2]);
			argc -= 2;
			argv += 2;
		} else if (argc >= 2 && !strncmp(argv[1], "-I", 2)) {
			Iflag = argv[1][2] == '2' ? 2 : 1;
			argc -= 1;
			argv += 1;
		} else if (argc >= 6 && !strcmp(argv[1], "-a1")) {
			Timadj[1].add = atof(argv[2]) * 1000;
			Timadj[1].mpy = atof(argv[3]);
			argc -= 3;
			argv += 3;
		} else if (argc >= 6 && !strcmp(argv[1], "-a2")) {
			Timadj[2].add = atof(argv[2]) * 1000;
			Timadj[2].mpy = atof(argv[3]);
			argc -= 3;
			argv += 3;
		} else if (argc >= 2 && !strncmp(argv[1], "-V", 2)) {
			Maxvvv += 1;
			argc -= 1;
			argv += 1;
		} else if (argc >= 2 && !strncmp(argv[1], "-W", 2)) {
			Maxvvv -= 1;
			argc -= 1;
			argv += 1;
		} else if (argc >= 2 && !strncmp(argv[1], "-S", 2)) {
			Supwrap = 1;
			argc -= 1;
			argv += 1;
		} else {
			break;
		}
	}
	if (argc != 3) {
		fprintf(stderr,
"Usage: subtitle_merge [-V] [-S] [-I[12]] [-C[12] #] [-a1 # #] [-a2 # #] Filename1 Filename2\n");
		fprintf(stderr,
"   Or: subtitle_merge [-a1 # #] [-C1 #] Filename1 none\n");
		fprintf(stderr,
"    -- -I#n option adds italics for text in n'th file.\n");
		fprintf(stderr,
"    -- -C#n option adds color for text in n'th file.\n");
		fprintf(stderr,
"    -- -a#n #Add #Mult option replaces timestanp #P in n'th file\n");
		fprintf(stderr,
"         with (P * Mult + Add), where all are in seconds\n");
		fprintf(stderr,
"    -- -V option increases vertical gap between concatenated subtitles\n");
		fprintf(stderr,
"    -- -B option increases vertical gap between concatenated subtitles\n");
		fprintf(stderr,
"    -- -S option suppresses line-wrap\n");
		exit(1);
	}
	p = Subt;
	for (fno = 1; fno < 3; fno++) {
		if (! strcmp(argv[fno], "none")) {
			Onlyonefile = 1;
			continue;
		}
		Firstinf = 1;
		The_file = fopen(argv[fno], "r");
		if (! The_file) {
			fprintf(stderr, "Unable to open '%s'\n", argv[fno]);
			exit(1);
		}
		/* Guess from filename whether it is srt or not */
		switch (lastc(argv[fno])) {
		case 'b': case 'B':
			Notsrt = 1;
			break;
		default:
			fprintf(stderr, "File *%s* -- assuming it is .srt\n",
				argv[fno]);
			/*FALLTHRU*/
		case 't': case 'T':
			Notsrt = 0;
			break;
		}
		while (bufp = jgets(Buff)) {
			if (Notsrt && Buff[0] == '[')
				continue;
			if (Notsrt && Buff[0] == '\0')
				continue;
			if (Notsrt) {
				pstop = strstr(bufp, ",");
				p->tstart = time2int(bufp, fno, 1);
				p->tstop  = time2int(pstop + 1, fno, 1);
				goto sscommo;
			} else if (pstop = strstr(bufp, "-->")) {
				p->tstart = time2int(bufp, fno, 0);
				p->tstop  = time2int(pstop + 4, fno, 0);
			sscommo:
#ifdef	VVBUGGY
				p->ktstart = p->tstart;
				p->ktstop  = p->tstop;
#endif
				p->subfno  = fno;
				while (1) {
					bufp = jgets(Buff);
					if (bufp == 0) {
						break;
					} else if (bufp[0] == '\0') {
						break;
					} else if (! strcmp(bufp, " ")) {
						break;
					}
					if (p->numline >= 9) {
						fprintf(stderr,
		"Subtitle longer than 9 lines.  Fix source, recompile.\n");
						exit(1);
					}
	/* Allocate 1 for the EOL null, and 1 for the strcat in fixuptext */
					p->stext[p->numline]
						= malloc(strlen(bufp) + 1 + 1);
					strcpy(p->stext[p->numline], bufp);
					/*
					 * Next added due to weirdness
					 * in Dr. Zhivago Thai subtitles.
					 */
					if (bufp[0] == '\r')
						++(p->stext[p->numline]);
					++(p->numline);
				}
				fixuptext(p);
				p += 1;
			}
		}
	}
	endp = p;

	/*
	 * The next may appear klugey, and weird
	 * and is NOT really necessary.
	 * Throw it away if it appears troublesome.
	 *
	 * (If the integer following is 1, it's already disabled.)
	 */
	if (0 || Onlyonefile)
		goto skipkluge;
	qsort(Subt, endp - Subt, sizeof *Subt, subcomp2);
	stopadj = adjcount = 0;
	for (q = Subt; q < endp - 1; q++) {
		adjf = (q+1)->subfno - q->subfno;
		if (!adjf)
			continue;
		tim0 = (q+1)->tstop - q->tstop;
		if (tim0 < 570) {
			adjcount += 1;
			stopadj += adjf == 1 ? tim0 : -tim0;
		}
	}
	stopadj /= adjcount;
	stopadj *= KLUGFACT;
	qsort(Subt, endp - Subt, sizeof *Subt, subcomp);
	startadj = adjcount = 0;
	for (q = Subt; q < endp - 1; q++) {
		adjf = (q+1)->subfno - q->subfno;
		if (!adjf)
			continue;
		tim0 = (q+1)->tstart - q->tstart;
		if (tim0 < 570) {
			adjcount += 1;
			startadj += adjf == 1 ? tim0 : -tim0;
		}
	}
	startadj /= adjcount;
	startadj *= KLUGFACT;
	for (q = Subt; q < endp; q++) {
		if (q->subfno == 1) {
			q->tstart += startadj;
			q->tstop += startadj;
		} else {
			q->tstart -= startadj;
			q->tstop -= startadj;
		}
	}
	skipkluge:
	/* End of unnecessary kluge code */

	qsort(Subt, endp - Subt, sizeof *Subt, subcomp);
	/* 30-9-20 :  Never suppress Boms */
	if (1 || Boms)
		printf("%c%c%c", 0xef, 0xbb, 0xbf);
	/*
	 * Eliminate small overlaps
	 */
	for (q = Subt; q < endp - 2; q++) {
		tim0 = q->tstop - (q+2)->tstart;
		if (tim0 < -2 || tim0 > 300)
			continue;
#ifdef	VVBUGGY
fprintf(stderr, "Overlap TWO %s %s --> %s %s --> ",
	int2time(q->ktstop),
	int2time((q+2)->ktstart),
	int2time(q->tstop),
	int2time((q+2)->tstart));
#endif

		(q+2)->tstart += tim0 / 2;
		q->tstop = (q+2)->tstart - 1;

#ifdef	VVBUGGY
fprintf(stderr, " %s %s\n",
	int2time(q->tstop),
	int2time((q+2)->tstart));
#endif

	}
	for (q = Subt; q < endp - 1; q++) {
		tim0 = q->tstop - (q+1)->tstart;
		if (tim0 < -2 || tim0 > 200)
			continue;
#ifdef	VVBUGGY
fprintf(stderr, "Overlap ONE %s %s --> %s %s --> ",
	int2time(q->ktstop),
	int2time((q+1)->ktstart),
	int2time(q->tstop),
	int2time((q+1)->tstart));
#endif

		(q+1)->tstart += tim0 / 2;
		q->tstop = (q+1)->tstart - 1;

#ifdef	VVBUGGY
fprintf(stderr, " %s %s\n",
	int2time(q->tstop),
	int2time((q+1)->tstart));
#endif

	}
	q = Subt;
#if	1
	/* Strip away lines with negative start times */
	for (; q < endp; q++)
		if (q->tstart > 0)
			break;
#endif

	for (cntr = 1; q < endp; cntr++) {
		if (0 && q < endp - 2)
				fprintf(stderr,
					"Top of main output loop:  %d (%d) %s %s %s\n",
					cntr, q - Subt,
					int2time(q->tstart),
					int2time((q+1)->tstart),
					int2time((q+2)->tstart));
		printf("\n%d\n%s --> ", cntr, int2time(q->tstart));
		if (q+1 < endp && q->tstop >= (q+1)->tstart) {
			tim1 = (q+1)->tstop;
			tim0 = (q+0)->tstop;
			if (q+2 < endp && tim1 >= (q+2)->tstart + 900) {
				/* q+1 covers both q and q+2 */
				q1 = q + 1;
				q2 = q + 2;
				timmed = (tim0 + (q+2)->tstart) / 2;
				tim0 = tim1;
				goto commo3;
			} else if (q+2 < endp && tim0 >= (q+2)->tstart) {
				/* q covers both q+1 and q+2 */
				q1 = q + 0;
				q2 = q + 2;
				timmed = (tim1 + (q+2)->tstart) / 2;
			commo3:
				printf("%s\n", int2time(timmed));
				printpair(q + 0, q + 1);
				cntr++;
				q += 1;
				ktim1 = tim1 = (q+1)->tstop;
				tim3 = q+2 < p ? (q+2)->tstart
						: 99999999;
				if (tim1 < tim0)
					tim1 = tim0;
				if (tim1 > tim3 - 5)
					tim1 = tim3 - 5;
#if	1
				if (tim1 - timmed - 1 < 50) {
					fprintf(stderr,
						"Subt with duration %d at %d\n",
						tim1 - timmed - 1, cntr);
					--cntr;
				} else
#endif
				{
				printf("\n%d\n%s --> ", cntr,
						int2time(timmed + 1));
				printf("%s\n", int2time(tim1));
				printpair(q1, q2);
				}
				if (ktim1 > tim1 + 400) {
					(q+1)->tstart = tim1 + 1;
					q += 1;
				} else {
					q += 2;
				}
			} else {
				q1 = q + 0;
				q2 = q + 1;
				tim3 = q+2 < p ? (q+2)->tstart
						: 99999999;
				if (tim1 < tim0)
					tim1 = tim0;
				if (tim1 > tim3 - 5)
					tim1 = tim3 - 5;
				printf("%s\n", int2time(tim1));
				printpair(q1, q2);
				q += 2;
			}
		} else {
			printf("%s\n", int2time(q->tstop));
			printsubs(q + 0);
			q += 1;
		}
	}
	printf("\n");
	exit(0);
}
